import pytest
import numpy as np
from numpy.testing import (assert_array_equal, assert_array_almost_equal_nulp,
                           assert_equal, assert_allclose)

from sklearn.datasets import make_regression
from sklearn.linear_model import Lasso
from sklearn.metrics import r2_score
from sklearn.linear_model._coordinate_descent import _alpha_grid
try:
    import pycasso
except ImportError:
    pycasso = None

from pyuoi import UoI_Lasso
from pyuoi.linear_model.lasso import PycLasso
from pyuoi.datasets import make_linear_regression


def test_variable_selection():
    """Test basic functionality of UoI_Lasso and that it finds right model"""

    X, y, w = make_regression(coef=True, random_state=1)
    lasso = UoI_Lasso()
    lasso.fit(X, y)
    true_coef = np.nonzero(w)[0]
    fit_coef = np.nonzero(lasso.coef_)[0]
    assert_array_equal(true_coef, fit_coef)
    assert_array_almost_equal_nulp(true_coef, fit_coef)


def test_estimation_score_usage():
    """Test the ability to change the estimation score in UoI Lasso."""

    methods = ('r2', 'AIC', 'AICc', 'BIC')
    X, y = make_regression(n_features=10, n_informative=3,
                           random_state=10)
    scores = []
    for method in methods:
        lasso = UoI_Lasso(estimation_score=method)
        assert_equal(lasso.estimation_score, method)
        lasso.fit(X, y)
        y_hat = lasso.predict(X)
        assert_equal(r2_score(y, y_hat), lasso.score(X, y))
        score = np.max(lasso.scores_)
        scores.append(score)
    assert_equal(len(np.unique(scores)), len(methods))


def test_set_random_state():
    """Tests whether random states are handled correctly."""
    X, y = make_regression(n_features=5, n_informative=3,
                           random_state=16, noise=.5)
    # same state
    l1log_0 = UoI_Lasso(random_state=13)
    l1log_1 = UoI_Lasso(random_state=13)
    l1log_0.fit(X, y)
    l1log_1.fit(X, y)
    assert_array_equal(l1log_0.coef_, l1log_1.coef_)

    # different state
    l1log_1 = UoI_Lasso(random_state=14)
    l1log_1.fit(X, y)
    assert not np.array_equal(l1log_0.coef_, l1log_1.coef_)

    # different state, not set
    l1log_0 = UoI_Lasso()
    l1log_1 = UoI_Lasso()
    l1log_0.fit(X, y)
    l1log_1.fit(X, y)
    assert not np.array_equal(l1log_0.coef_, l1log_1.coef_)


def test_uoi_lasso_toy():
    """Test UoI Lasso on a toy example."""

    X = np.array([
        [-1, 2],
        [4, 1],
        [1, 3],
        [4, 3],
        [8, 11]], dtype=float)
    beta = np.array([1, 4], dtype=float)
    y = np.dot(X, beta)

    # choose selection_frac to be slightly smaller to ensure that we get
    # good test sets

    # Also test both choices of solver
    lasso = UoI_Lasso(
        fit_intercept=False,
        selection_frac=0.75,
        estimation_frac=0.75,
        solver='cd'
    )
    lasso.fit(X, y)
    lasso.fit(X, y, verbose=True)

    assert_allclose(lasso.coef_, beta)

    if pycasso is not None:
        lasso = UoI_Lasso(
            fit_intercept=False,
            selection_frac=0.75,
            estimation_frac=0.75,
            solver='pyc'
        )
        lasso.fit(X, y)

        assert_allclose(lasso.coef_, beta)


def test_uoi_lasso_estimation_shape_match():
    """Test UoI Lasso on a toy example."""
    n_samples = 40
    n_features = 10
    X, y = make_regression(n_samples=n_samples,
                           n_features=n_features,
                           n_informative=5,
                           random_state=1)

    lasso = UoI_Lasso()
    lasso.fit(X, y)
    with pytest.raises(ValueError, match='Targets and predictions are ' +
                       'not the same shape.'):
        support = np.ones(n_features).astype(bool)
        boot_idxs = [np.arange(n_samples)] * 2
        lasso.coef_ = np.random.randn(2, n_features)
        lasso._score_predictions('r2', lasso, X, y,
                                 support, boot_idxs)

    with pytest.raises(ValueError, match='y should either have'):
        support = np.ones(n_features).astype(bool)
        boot_idxs = [np.arange(n_samples)] * 2
        lasso._score_predictions('r2', lasso, X, y[:, np.newaxis, np.newaxis],
                                 support, boot_idxs)


def test_uoi_lasso_fit_shape_match():
    """Test UoI Lasso on a toy example."""

    n_samples = 40
    n_features = 10
    X, y = make_regression(n_samples=n_samples,
                           n_features=n_features,
                           n_informative=5,
                           random_state=1)

    lasso = UoI_Lasso()

    # Check that second axis gets squeezed
    lasso.fit(X, y[:, np.newaxis])

    # Check that second axis gets squeezed
    message = 'y should either have shape'
    with pytest.raises(ValueError, match=message):
        lasso.fit(X, np.tile(y[:, np.newaxis], (1, 2)))
    with pytest.raises(ValueError, match=message):
        lasso.fit(X, y[:, np.newaxis, np.newaxis])


def test_get_reg_params():
    """Tests whether get_reg_params works correctly for UoI Lasso."""

    X = np.array([
        [-1, 2],
        [0, 1],
        [1, 3],
        [4, 3]])
    y = np.array([7, 4, 13, 16])

    # calculate regularization parameters manually
    alpha_max = np.max(np.dot(X.T, y) / 4)
    alphas = [{'alpha': alpha_max}, {'alpha': alpha_max / 10.}]

    # calculate regularization parameters with UoI_Lasso object
    lasso = UoI_Lasso(
        n_lambdas=2,
        fit_intercept=False,
        eps=0.1)
    reg_params = lasso.get_reg_params(X, y)

    # check each regularization parameter and key
    for estimate, true in zip(reg_params, alphas):
        assert estimate.keys() == true.keys()
        assert_allclose(list(estimate.values()), list(true.values()))


def test_intercept_and_coefs_no_selection():
    """Test that UoI Lasso properly calculates the intercept with and without
    standardization."""
    # create line model
    X, y, beta, intercept = make_linear_regression(
        n_samples=500,
        n_features=2,
        n_informative=2,
        snr=10.,
        include_intercept=True,
        random_state=2332)

    # without standardization
    lasso = UoI_Lasso(
        standardize=False,
        fit_intercept=True
    )
    lasso.fit(X, y)
    assert_allclose(lasso.intercept_, intercept, rtol=0.25)
    assert_allclose(lasso.coef_, beta, rtol=0.25)

    # with standardization
    lasso = UoI_Lasso(
        standardize=True,
        fit_intercept=True
    )
    lasso.fit(X, y)
    assert_allclose(lasso.intercept_, intercept, rtol=0.25)
    assert_allclose(lasso.coef_, beta, rtol=0.25)


def test_lasso_selection_sweep():
    """Tests uoi_selection_sweep for UoI_Lasso."""

    # toy data
    X = np.array([
        [-1, 2, 3],
        [4, 1, -7],
        [1, 3, 1],
        [4, 3, 12],
        [8, 11, 2]], dtype=float)
    beta = np.array([1, 4, 2], dtype=float)
    y = np.dot(X, beta)

    # toy regularization
    reg_param_values = [{'alpha': 1.0}, {'alpha': 2.0}]
    lasso = UoI_Lasso(fit_intercept=True, warm_start=False)
    lasso1 = Lasso(alpha=1.0, fit_intercept=True, max_iter=lasso.max_iter)
    lasso2 = Lasso(alpha=2.0, fit_intercept=True, max_iter=lasso.max_iter)
    lasso.output_dim = 1

    coefs = lasso.uoi_selection_sweep(X, y, reg_param_values)
    lasso1.fit(X, y)
    lasso2.fit(X, y)

    assert np.allclose(coefs[0], lasso1.coef_)
    assert np.allclose(coefs[1], lasso2.coef_)


def test_fit_intercept():
    """Tests whether `include_intercept` in passed through to the linear models.
    """
    lasso = UoI_Lasso(fit_intercept=True)
    assert lasso._selection_lm.fit_intercept
    assert lasso._estimation_lm.fit_intercept

    lasso = UoI_Lasso(fit_intercept=False)
    assert not lasso._selection_lm.fit_intercept
    assert not lasso._estimation_lm.fit_intercept


@pytest.mark.skipif(pycasso is None, reason='pycasso not installed')
def test_choice_of_solver():
    '''Tests whether one can correctly switch between solvers in UoI Lasso'''

    uoi1 = UoI_Lasso(solver='cd')
    assert(isinstance(uoi1._selection_lm, Lasso))

    uoi2 = UoI_Lasso(solver='pyc')
    assert(isinstance(uoi2._selection_lm, PycLasso))


@pytest.mark.skipif(pycasso is not None, reason='pycasso is installed')
def test_pycasso_error():
    """Tests whether an error is raised if pycasso is not installed.
    """

    with pytest.raises(ImportError):
        uoi2 = UoI_Lasso(solver='pyc')
        assert(isinstance(uoi2._selection_lm, PycLasso))


@pytest.mark.skipif(pycasso is None, reason='pycasso not installed')
def test_pyclasso():
    """Tests whether the PycLasso class is working"""

    pyclasso = PycLasso(fit_intercept=False, max_iter=1000)

    # Test that we can set params correctly
    pyclasso.set_params(fit_intercept=True)
    assert(pyclasso.fit_intercept)
    pyclasso.set_params(max_iter=500)
    assert(pyclasso.max_iter == 500)
    pyclasso.set_params(alphas=np.arange(100))
    assert(np.array_equal(pyclasso.alphas, np.arange(100)))

    # Test that spurious parameters are rejected
    try:
        pyclasso.set_params(blah=5)
        Exception('No exception thrown!')
    except ValueError:
        pass
    finally:
        Exception('Unexpected Exception raised')

    # Tests against a toy problem
    X = np.array([
        [-1, 2, 3],
        [4, 1, -7],
        [1, 3, 1],
        [4, 3, 12],
        [8, 11, 2]], dtype=float)
    beta = np.array([1, 4, 2], dtype=float)
    y = np.dot(X, beta)

    alphas = _alpha_grid(X, y)
    pyclasso.set_params(alphas=alphas, fit_intercept=False)
    pyclasso.fit(X, y)
    assert(np.array_equal(pyclasso.coef_.shape, (100, 3)))
    y_pred = pyclasso.predict(X)
    scores = np.array([r2_score(y, y_pred[:, j]) for j in range(100)])
    assert(np.allclose(1, max(scores)))


def test_lass_bad_est_score():
    """Test that UoI Lasso raises an error when given a bad
    estimation_score value.
    """
    X = np.random.randn(20, 5)
    y = np.random.randn(20)

    with pytest.raises(ValueError):
        UoI_Lasso(estimation_score='z',
                  n_boots_sel=10, n_boots_est=10).fit(X, y)



"""Union of Intersection models with matrix decomposition."""
from .CUR import UoI_CUR
from .CUR import CUR
from .NMF import UoI_NMF
from .NMF import UoI_NMF_Base

__all__ = ["UoI_CUR",
           "CUR",
           "UoI_NMF",
           "UoI_NMF_Base"]



import numpy as np
import sys
import logging


def softmax(y, axis=-1):
    """Calculates the softmax distribution.

    Parameters
    ----------
    y : ndarray
        Log-probabilities.
    """

    yp = y - y.max(axis=axis, keepdims=True)
    epy = np.exp(yp)
    return epy / np.sum(epy, axis=axis, keepdims=True)


def sigmoid(x):
    """Calculates the bernoulli distribution.

    Parameters
    ----------
    x : ndarray
        Log-probabilities.
    """
    return np.exp(-np.logaddexp(0, -x))


def log_likelihood_glm(model, y_true, y_pred):
    """Calculates the log-likelihood of a generalized linear model given the
    true response variables and the "predicted" response variables. The
    "predicted" response variable varies by the specific generalized linear
    model under consideration.

    Parameters
    ----------
    model : string
        The generalized linear model to calculate the log-likelihood for.
    y_true : nd-array, shape (n_samples,)
        Array of true response values.
    y_pred : nd-array, shape (n_samples,)
        Array of predicted response values (conditional mean).

    Returns
    -------
    ll : float
        The log-likelihood.
    """
    if model == 'normal':
        # this log-likelihood is calculated under the assumption that the
        # variance is the value that maximizes the log-likelihood
        rss = (y_true - y_pred)**2
        n_samples = y_true.size
        ll = -n_samples / 2 * (1 + np.log(np.mean(rss)))
    elif model == 'poisson':
        if not np.any(y_pred):
            if np.any(y_true):
                ll = -np.inf
            else:
                ll = 0.
        else:
            ll = np.mean(y_true * np.log(y_pred) - y_pred)
    else:
        raise ValueError('Model is not available.')
    return ll


def BIC(ll, n_features, n_samples):
    """Calculates the Bayesian Information Criterion.

    Parameters
    ----------
    ll : float
        The log-likelihood of the model.
    n_features : int
        The number of features used in the model.
    n_samples : int
        The number of samples in the dataset being tested.

    Returns
    -------
    BIC : float
        Bayesian Information Criterion
    """
    BIC = n_features * np.log(n_samples) - 2 * ll
    return BIC


def AIC(ll, n_features):
    """Calculates the Akaike Information Criterion.

    Parameters
    ----------
    ll : float
        The log-likelihood of the model.
    n_features : int
        The number of features used in the model.
    n_samples : int
        The number of samples in the dataset being tested.

    Returns
    -------
    AIC : float
        Akaike Information Criterion
    """

    AIC = 2 * n_features - 2 * ll
    return AIC


def AICc(ll, n_features, n_samples):
    """Calculate the corrected Akaike Information Criterion. This criterion is
    useful in cases when the number of samples is small.

    If the number of features is equal to the number of samples plus one, then
    the AIC is returned (the AICc is undefined in this case).

    Parameters
    ----------
    ll : float
        The log-likelihood of the model.
    n_features : int
        The number of features used in the model.
    n_samples : int
        The number of samples in the dataset being tested.

    Returns
    -------
    AIC : float
        Akaike Information Criterion
    """
    AICc = AIC(ll, n_features)
    if n_samples > (n_features + 1):
        AICc += 2 * (n_features**2 + n_features) / (n_samples - n_features - 1)
    return AICc


def check_logger(logger, name='uoi', comm=None):
    ret = logger
    if ret is None:
        if comm is not None and comm.Get_size() > 1:
            r, s = comm.Get_rank(), comm.Get_size()
            name += " " + str(r).rjust(int(np.log10(s)) + 1)

        ret = logging.getLogger(name=name)
        handler = logging.StreamHandler(sys.stdout)

        fmt = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'

        handler.setFormatter(logging.Formatter(fmt))
        ret.addHandler(handler)
    return ret



from setuptools import setup, find_packages, Extension
from setuptools.command.build_ext import build_ext
from distutils.ccompiler import get_default_compiler
# To use a consistent encoding
from codecs import open
from os import path

import numpy as np


here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()
with open(path.join(here, 'requirements.txt'), encoding='utf-8') as f:
    requirements = f.read().splitlines()
with open(path.join(here, 'requirements-dev.txt'), encoding='utf-8') as f:
    dev_requirements = f.read().splitlines()
    dev_requirements = dev_requirements[1:] # Throw away the first line which is not a package.

# Prepare lbfgs
from Cython.Build import cythonize

class custom_build_ext(build_ext):
    def finalize_options(self):
        build_ext.finalize_options(self)
        if self.compiler is None:
            compiler = get_default_compiler()
        else:
            compiler = self.compiler

        if compiler == 'msvc':
            include_dirs.append('compat/win32')

include_dirs = ['liblbfgs', np.get_include()]

ext_modules = cythonize(
    [Extension('pyuoi.lbfgs._lowlevel',
               ['pyuoi/lbfgs/_lowlevel.pyx', 'liblbfgs/lbfgs.c'],
               include_dirs=include_dirs)])


setup(
    name='pyuoi',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version='1.1.0',

    description='The Union of Intersections framework in Python.',
    long_description=long_description,
    long_description_content_type="text/markdown",


    # Author details
    author='',
    author_email='',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: BSD License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],

    # What does your project relate to?
    keywords='UoI',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(),
    package_data={'pyuoi': ['data/*']},

    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=requirements,

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        'perf': ['mpi4py', 'pycasso'],
        'dev': dev_requirements
    },

    url='https://github.com/BouchardLab/pyuoi',
    ext_modules=ext_modules,
    cmdclass={'build_ext': custom_build_ext}


    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    # entry_points={
    #    'console_scripts': [
    #        'sample=sample:main',
    #    ],
    # },
)



import h5py, pytest
import numpy as np

from numpy.testing import assert_array_equal
try:
    from mpi4py import MPI
except ImportError:
    MPI = None

from pyuoi.mpi_utils import (Bcast_from_root, Gatherv_rows,
                             load_data_MPI)


@pytest.mark.skipif(MPI is None, reason='MPI not installed.')
def test_load_data_MPI(tmpdir):
    """Tests loading data from an HDF5 file into all ranks.
    """
    comm = MPI.COMM_WORLD
    rank = comm.rank
    root = 0
    X = np.random.randn(5, 10)
    y = np.random.randint(5, size=5)

    fname = tmpdir.join('temp.h5')
    if rank == root:
        with h5py.File(str(fname), 'w') as f:
            f.create_dataset('X', data=X)
            f.create_dataset('Xp', data=X)
            f.create_dataset('y', data=y)
            f.create_dataset('yp', data=y)

    # Default keys
    X_load, y_load = load_data_MPI(fname)
    if rank == root:
        assert_array_equal(X, X_load)
        assert_array_equal(y, y_load)

    # Set keys
    X_load, y_load = load_data_MPI(fname,
                                   X_key='Xp',
                                   y_key='yp')
    if rank == root:
        assert_array_equal(X, X_load)
        assert_array_equal(y, y_load)


@pytest.mark.skipif(MPI is None, reason='MPI not installed.')
def test_Bcast_from_root():
    """Test the Bcast_from_root function for broadcasting
    an array from root to all ranks.
    """
    comm = MPI.COMM_WORLD
    root = 0

    dims = [2, 3, 5]

    for dtype in [int, float]:
        for ndim in range(1, 4):
            my_dim = dims[:ndim]
            X = None
            if comm.rank == root:
                X = np.arange(np.prod(my_dim), dtype=dtype)
                X = X.reshape(my_dim)
            X = Bcast_from_root(X, comm, root)
            Xp = np.arange(np.prod(my_dim), dtype=dtype)
            Xp = Xp.reshape(my_dim)
            assert_array_equal(X, Xp)
            assert X.dtype == dtype
            assert X.ndim == len(my_dim)


@pytest.mark.skipif(MPI is None, reason='MPI not installed.')
def test_Gatherv_rows():
    """Test the Gatherv_rows function for Gathering and
    concatenating ndarrys along their first axes to root.
    """
    comm = MPI.COMM_WORLD
    root = 0
    rank = comm.rank
    size = comm.size

    for dtype in [int, float]:
        # Multiple rows per rank
        X = np.arange(151 * 3, dtype=dtype).reshape(151, 3)
        my_rows = np.array_split(X, size)[rank]
        Xp = Gatherv_rows(my_rows, comm, root)
        if rank == root:
            assert_array_equal(X, Xp)
            assert Xp.dtype == dtype

        # Fewer rows than ranks
        X = np.arange(2 * 3, dtype=dtype).reshape(2, 3)
        my_rows = np.array_split(X, size)[rank]
        Xp = Gatherv_rows(my_rows, comm, root)
        if rank == root:
            assert_array_equal(X, Xp)
            assert Xp.dtype == dtype

        # Multiple rows per rank, 3d
        X = np.arange(151 * 2 * 3, dtype=dtype).reshape(151, 2, 3)
        my_rows = np.array_split(X, size)[rank]
        Xp = Gatherv_rows(my_rows, comm, root)
        if rank == root:
            assert_array_equal(X, Xp)
            assert Xp.dtype == dtype

        # Fewer rows than ranks, 3d
        X = np.arange(2 * 3 * 5, dtype=dtype).reshape(2, 3, 5)
        my_rows = np.array_split(X, size)[rank]
        Xp = Gatherv_rows(my_rows, comm, root)
        if rank == root:
            assert_array_equal(X, Xp)
            assert Xp.dtype == dtype


@pytest.mark.skipif(MPI is None, reason='MPI not installed.')
def test_Gatherv_random_rows():
    """Test Gatherv_rows for gathering ndarrays with random
    shapes along their first axis
    """

    comm = MPI.COMM_WORLD
    root = 0
    rank = comm.rank

    data = np.random.normal(size=(np.random.randint(1, 10), 1000))
    sizes = comm.gather(data.shape[0], root=root)
    data = Gatherv_rows(data, comm, root)

    if rank == root:
        assert(data.shape[0] == np.sum(sizes))



import numpy as np

from .base import AbstractUoILinearRegressor

from sklearn.linear_model import LinearRegression
from sklearn.linear_model._coordinate_descent import _alpha_grid
from sklearn.linear_model import ElasticNet


class UoI_ElasticNet(AbstractUoILinearRegressor, LinearRegression):
    r"""UoI\ :sub:`ElasticNet` solver.

    Parameters
    ----------
    n_boots_sel : int
        The number of data bootstraps to use in the selection module.
        Increasing this number will make selection more strict.
    n_boots_est : int
        The number of data bootstraps to use in the estimation module.
        Increasing this number will relax selection and decrease variance.
    selection_frac : float
        The fraction of the dataset to use for training in each resampled
        bootstrap, during the selection module. Small values of this parameter
        imply larger "perturbations" to the dataset.
    estimation_frac : float
        The fraction of the dataset to use for training in each resampled
        bootstrap, during the estimation module. The remaining data is used
        to obtain validation scores. Small values of this parameters imply
        larger "perturbations" to the dataset. IGNORED - Leaving this here
        to double check later
    n_lambdas : int
        The number of regularization values to use for selection.
    alphas : list or ndarray
        The parameter that trades off L1 versus L2 regularization for a given
        lambda.
    stability_selection : int, float, or array-like
        If int, treated as the number of bootstraps that a feature must
        appear in to guarantee placement in selection profile. If float,
        must be between 0 and 1, and is instead the proportion of
        bootstraps. If array-like, must consist of either ints or floats
        between 0 and 1. In this case, each entry in the array-like object
        will act as a separate threshold for placement in the selection
        profile.
    estimation_score : string, "r2" | "AIC" | "AICc" | "BIC"
        Objective used to choose the best estimates per bootstrap.
    estimation_target : string, "train" | "test"
        Decide whether to assess the estimation_score on the train
        or test data across each bootstrap. By default, a sensible
        choice is made based on the chosen estimation_score.
    warm_start : bool
        When set to ``True``, reuse the solution of the previous call to fit as
        initialization, otherwise, just erase the previous solution
    eps : float
        Length of the lasso path. ``eps=1e-3`` means that
        ``alpha_min / alpha_max = 1e-3``.
    copy_X : bool
        If ``True``, X will be copied; else, it may be overwritten.
    fit_intercept : bool
        Whether to calculate the intercept for this model. If set
        to False, no intercept will be used in calculations
        (e.g. data is expected to be already centered).
    standardize : bool
        If True, the regressors X will be standardized before regression by
        subtracting the mean and dividing by their standard deviations.
    max_iter : int
        Maximum number of iterations for iterative fitting methods.
    random_state : int, RandomState instance, or None
        The seed of the pseudo random number generator that selects a random
        feature to update.  If int, random_state is the seed used by the random
        number generator; If RandomState instance, random_state is the random
        number generator; If None, the random number generator is the
        RandomState instance used by `np.random`.
    comm : MPI communicator
        If passed, the selection and estimation steps are parallelized.
    logger : Logger
        The logger to use for messages when ``verbose=True`` in ``fit``.
        If *None* is passed, a logger that writes to ``sys.stdout`` will be
        used.

    Attributes
    ----------
    coef_ : array, shape (n_features,) or (n_targets, n_features)
        Estimated coefficients for the linear regression problem.
    intercept_ : float
        Independent term in the linear model.
    supports_ : ndarray, shape (n_supports, n_features)
        Boolean array indicating whether a given regressor (column) is selected
        for estimation for a given regularization parameter value (row).
    """
    def __init__(self, n_boots_sel=24, n_boots_est=24, selection_frac=0.9,
                 estimation_frac=0.9, n_lambdas=48,
                 alphas=np.array([0.5]), stability_selection=1.,
                 estimation_score='r2', estimation_target=None,
                 warm_start=True, eps=1e-3, copy_X=True,
                 fit_intercept=True, standardize=True,
                 max_iter=1000, random_state=None, comm=None, logger=None):
        super(UoI_ElasticNet, self).__init__(
            n_boots_sel=n_boots_sel,
            n_boots_est=n_boots_est,
            selection_frac=selection_frac,
            estimation_frac=estimation_frac,
            stability_selection=stability_selection,
            estimation_score=estimation_score,
            estimation_target=estimation_target,
            copy_X=copy_X,
            fit_intercept=fit_intercept,
            standardize=standardize,
            random_state=random_state,
            comm=comm,
            max_iter=max_iter,
            logger=logger)
        self.n_lambdas = n_lambdas
        self.alphas = alphas
        self.n_alphas = len(alphas)
        self.warm_start = warm_start
        self.eps = eps
        self.lambdas = None
        self._selection_lm = ElasticNet(
            fit_intercept=fit_intercept,
            max_iter=max_iter,
            copy_X=copy_X,
            warm_start=warm_start,
            random_state=random_state)
        self._estimation_lm = LinearRegression(fit_intercept=fit_intercept)

    def get_reg_params(self, X, y):
        r"""Calculates the regularization parameters (alpha and lambda) to be
        used for the provided data.

        Note that the Elastic Net penalty is given by

        .. math::
           \frac{1}{2\ \text{n_samples}} ||y - Xb||^2_2
           + \lambda (\alpha |b|_1 + 0.5 (1 - \alpha) |b|^2_2)

        where lambda and alpha are regularization parameters.

        ``scikit-learn`` does not use these names. Instead, ``scitkit-learn``
        denotes alpha by 'l1_ratio' and lambda by 'alpha'.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The design matrix.

        y : array-like, shape (n_samples)
            The response vector.

        Returns
        -------
        reg_params : a list of dictionaries
            A list containing dictionaries with the value of each
            (lambda, alpha) describing the type of regularization to impose.
            The keys adhere to scikit-learn's terminology (lambda->alpha,
            alpha->l1_ratio). This allows easy passing into the ElasticNet
            object.
        """
        if self.lambdas is None:
            self.lambdas = np.zeros((self.n_alphas, self.n_lambdas))
            # a set of lambdas are generated for each alpha value (l1_ratio in
            # sci-kit learn parlance)
            for alpha_idx, alpha in enumerate(self.alphas):
                self.lambdas[alpha_idx, :] = _alpha_grid(
                    X=X, y=y,
                    l1_ratio=alpha,
                    fit_intercept=self.fit_intercept,
                    eps=self.eps,
                    n_alphas=self.n_lambdas)

        # place the regularization parameters into a list of dictionaries
        reg_params = list()
        for alpha_idx, alpha in enumerate(self.alphas):
            for lamb_idx, lamb in enumerate(self.lambdas[alpha_idx]):
                # reset the regularization parameter
                reg_params.append(dict(alpha=lamb, l1_ratio=alpha))

        return reg_params



"""Union of Intersection models with linear selection and estimation.

Provides both abstract base classes for creating user-defined UoI models
and several concrete implementations.
"""
from .base import (AbstractUoILinearModel, AbstractUoILinearRegressor,
                   AbstractUoIGeneralizedLinearRegressor)
from .lasso import UoI_Lasso
from .elasticnet import UoI_ElasticNet
from .logistic import MaskedCoefLogisticRegression, UoI_L1Logistic
from .poisson import Poisson, UoI_Poisson

__all__ = ["AbstractUoILinearModel",
           "AbstractUoILinearRegressor",
           "AbstractUoIGeneralizedLinearRegressor",
           "MaskedCoefLogisticRegression",
           "UoI_L1Logistic",
           "UoI_Lasso",
           "UoI_ElasticNet",
           "Poisson",
           "UoI_Poisson"]
